/* See LICENSE file for copyright and license details. */

#include <locale.h> /* setlocale(3), LC_CTYPE */
#include <stdio.h> /* fprintf(3), stderr */
#include <sysexits.h> /* EX_USAGE */
#include <unistd.h> /* getopt(3) */
#include <X11/Xlib.h> /* XSupportsLocale(3), Display, Window */
#include <X11/Xutil.h>
#ifdef XINERAMA
#	include <X11/extensions/Xinerama.h>
#endif
#include <X11/Xft/Xft.h> /* XftColor */
#include "drw.h" /* drw_clr_create(3), drw_create(3), drw_fontset_create(3),
                  * drw_fontset_getwidth(3), drw_free(3), drw_map(3),
                  * drw_resize(3), drw_setscheme(3), drw_rect(3), drw_text(3),
                  * Drw */

/* chars to be allocated for the text displayed in the bar */
#define BUFSIZE 1024

#define MIN(A, B) ((A) < (B) ? (A) : (B))
#define MAX(A, B) ((A) > (B) ? (A) : (B))

static XClassHint ch = {"dmenubar", "dmenubar"};

#define bool char
static struct {
	bool topbar;       /* -b */
	char *bg_hex;      /* -B */
	bool break_on_eof; /* -e */
	char *fg_hex;      /* -F */
	int mon;           /* -m */
	char *embed;       /* -w */
} args;
#undef bool

#define topbar_default 1
static char *bg_hex_default = "#005577";
#define break_on_eof_default 0
static char *fg_hex_default = "#EEEEEE";
#define mon_default -1
#define embed_default NULL

/* default X11 font or font set */
static const char *fonts[] = { /* drw_fontset_create(3) wants const char ** */
	"monospace:size=10" /* -f overrides this */
};

int
main(int argc, char *argv[]){
	int bh;
	XftColor bg;
	int c;
	Display *dpy;
	Drw *drw;
	unsigned int du;
	Window dw;
	Window *dws;
	XftColor fg;
	int i;
	int j;
	int lrpad; /* sum of left and right padding */
	int mh;
	int mw;
	extern char *optarg;
	Window parentwin;
	Window root;
	int screen;
	XSetWindowAttributes swa;
	char text[BUFSIZE];
	Window w;
	XWindowAttributes wa;
	Window win;
	int x;
	int y;
#ifdef XINERAMA
	XineramaScreenInfo *info;
	Window pw;
	int a, di, n, area = 0;
#endif

	args.topbar = topbar_default;
	args.bg_hex = bg_hex_default;
	args.break_on_eof = break_on_eof_default;
	args.fg_hex = fg_hex_default;
	args.mon = mon_default;
	args.embed = embed_default;
	while((c = getopt(argc, argv, ":bf:m:B:F:w:")) != -1)
		switch(c){
		case 'b':
			args.topbar = !args.topbar;
			break;
		case 'e':
			args.break_on_eof = !args.break_on_eof;
			break;
		case 'B': if(optarg){
			args.bg_hex = optarg;
			break;
		}case 'f': if(optarg){
			fonts[0] = optarg;
			break;
		}case 'F': if(optarg){
			args.fg_hex = optarg;
			break;
		}case 'm': if(optarg){
			args.mon = atoi(optarg);
			break;
		}case 'w': if(optarg){
			args.embed = optarg;
			break;
		}default: /* optarg==0 falls through */
			fprintf(stderr,
				"Usage: %s (-be) (-B [background color])"
				"(-f [font]) (-F foreground color)\n"
				"\t(-m [monitor]) (-w [windowid)\n",
				argv[0]);
			return EX_USAGE;
		}

	if(setlocale(LC_CTYPE, "") == NULL || !XSupportsLocale())
		fprintf(stderr, "%s: No locale support.\n", argv[0]);

	if((dpy = XOpenDisplay(NULL)) == NULL){
		fprintf(stderr, "%s: Cannot open X display.\n", argv[0]);
		return EX_UNAVAILABLE;
	}
	screen = DefaultScreen(dpy);
	root = RootWindow(dpy, screen);
	if (args.embed == NULL || !(parentwin = strtol(args.embed, NULL, 0)))
		parentwin = root;
	if (!XGetWindowAttributes(dpy, parentwin, &wa)){
		fprintf(stderr, "%s: Could not get embedding window "
			"attributes: 0x%lx", argv[0], parentwin);
		return EX_UNAVAILABLE;
	}
	drw = drw_create(dpy, screen, root, wa.width, wa.height);
	if (!drw_fontset_create(drw, fonts, (sizeof fonts) / (sizeof *fonts))){
		fprintf(stderr, "%s: No fonts could be loaded.\n", argv[0]);
		return EX_UNAVAILABLE;
	}
	lrpad = drw->fonts->h;

#ifdef __OpenBSD__
	if(pledge("stdio rpath", NULL) == -1){
		fprintf(stderr, "%s: Could not pledge(2).\n", argv[0]);
		return EX_UNAVAILABLE;
	}
#endif

	drw_clr_create(drw, &bg, args.bg_hex);
	drw_clr_create(drw, &fg, args.fg_hex);

	/* calculate menu geometry */
	bh = drw->fonts->h + 2;
	mh = bh;
#ifdef XINERAMA
#	define INTERSECT(x, y, w, h, r) (\
		MAX(0, \
			MIN((x) + (w), (r).x_org + (r).width) \
			- MAX((x), (r).x_org)) \
		* MAX(0, \
			MIN((y) + (h), (r).y_org + (r).height) \
			- MAX((y), (r).y_org))\
	)
	i = 0;
	if(parentwin == root && (info = XineramaQueryScreens(dpy, &n))){
		XGetInputFocus(dpy, &w, &di);
		if(args.mon >= 0 && args.mon < n)
			i = args.mon;
		else if(w != root && w != PointerRoot && w != None){
			/* find top-level window containing current input focus
			 */
			do{
				if(		XQueryTree(dpy, (pw = w), &dw,
							&w, &dws, &du)
						&& dws)
					XFree(dws);
			}while(w != root && w != pw);
			/* find xinerama screen with which the window
			 * intersects most */
			if(XGetWindowAttributes(dpy, pw, &wa))
				for(j = 0; j < n; j++)
					if(		(a = INTERSECT(wa.x,
								wa.y, wa.width,
								wa.height,
								info[j]))
							> area){
						area = a;
						i = j;
					}
		}
		/* no focused window is on screen, so use pointer location instead */
		if(		args.mon < 0
				&& !area
				&& XQueryPointer(dpy, root, &dw, &dw, &x, &y,
					&di, &di, &du))
			for(i = 0; i < n; i++)
				if(INTERSECT(x, y, 1, 1, info[i]))
					break;

		x = info[i].x_org;
		y = info[i].y_org + (args.topbar ? 0 : info[i].height - mh);
		mw = info[i].width;
		XFree(info);
	} else
#endif /* ifdef XINERAMA */
	{
		if(!XGetWindowAttributes(dpy, parentwin, &wa)){
			fprintf(stderr, 
				"%s: could not get embedding window"
				"attributes: 0x%lx\n",
				argv[0], parentwin);
			return EX_UNAVAILABLE; /* appropriate sysexit? */
		}
		x = 0;
		y = args.topbar ? 0 : wa.height - mh;
		mw = wa.width;
	}

	/* create menu window */
	swa.override_redirect = True;
	swa.background_pixel = bg.pixel;
	swa.event_mask = ExposureMask | KeyPressMask | VisibilityChangeMask;
	win = XCreateWindow(dpy, parentwin, x, y, mw, mh, 0,
	                    CopyFromParent, CopyFromParent, CopyFromParent,
	                    CWOverrideRedirect | CWBackPixel | CWEventMask, &swa);
	XSetClassHint(dpy, win, &ch);


	XMapRaised(dpy, win);
	if (args.embed != NULL) {
		XSelectInput(dpy, parentwin, FocusChangeMask | SubstructureNotifyMask);
		if (XQueryTree(dpy, parentwin, &dw, &w, &dws, &du) && dws) {
			for (i = 0; i < du && dws[i] != win; ++i)
				XSelectInput(dpy, dws[i], FocusChangeMask);
			XFree(dws);
		}
	}
	drw_resize(drw, mw, mh);
	drw_setscheme(drw, fg, bg);

	text[0] = '\0';
	do{
		/* trim newline */
		for(i = 0; text[i] != '\0' && text[i] != '\n';);
		if(text[i] == '\n')
			text[i] = '\0';

		drw_rect(drw, 0, 0, mw, mh, 1, 1);

		if(*text != '\0')
			drw_text(drw, x, 0,
				*text != '\0'
					? (drw_fontset_getwidth(drw, text)
						+ lrpad) - lrpad / 4
					: 0,
				bh, lrpad / 2, text, 0);

		drw_map(drw, win, 0, 0, mw, mh);
	}while(
		fgets(text, (sizeof text) / (sizeof *text), stdin) != NULL
		|| !args.break_on_eof
	);

	drw_free(drw);
	XSync(dpy, False);
	XCloseDisplay(dpy);

	return 0;
}
